use core::convert::{TryFrom, TryInto};
use core::fmt::Debug;
use std::collections::HashMap;
use std::path::Path;

use futures_util::future::try_join_all;
use hyper::client::connect::Connect;
use icinga_api::api::object_type::ObjectType;
use icinga_api::{api, icinga};
use tracing::{debug, error};

use crate::{format, fs, io, Arc, String};

#[tracing::instrument]
pub async fn do_actions<Connector>(
    service: &Arc<icinga::Service<Connector>>,
    parse_result: i2api_utils::parsing::ParseResult,
    author: &str,
    channel: &str,
    sender: &crate::sender::Sender,
) -> Result<(), icinga_api::utils::Error>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    use i2api_utils::parsing;

    let service_name = service.name();
    let service = Arc::clone(service);
    let action_handle = match parse_result {
        parsing::ParseResult::Highlight | parsing::ParseResult::Help => unreachable!(),
        parsing::ParseResult::Ack(p_res) => acknowledge(p_res, service, author)?,
        parsing::ParseResult::AckExp(p_res) => acknowledge_exp(p_res, service, author)?,
        parsing::ParseResult::Dt(p_res) => downtime(p_res, service, author)?,
        parsing::ParseResult::DtAt(p_res) => downtime_at(p_res, service, author)?,
        parsing::ParseResult::Com(p_res) => comment(p_res, service, author)?,
        parsing::ParseResult::Notif(active) => set_notification(active, service),
        parsing::ParseResult::Rm(p_res) => {
            let join_handle = remove(p_res, service, author)?;
            if let Ok(handles) = join_handle.await {
                for res in handles {
                    match res {
                        Ok(action_result) => {
                            if action_result.results.is_empty() {
                                sender
                                    .send_notice(
                                        &channel,
                                        format!("{author}: Nothing to do -- {service_name}"),
                                    )
                                    .await
                                    .unwrap();
                            }

                            for res in action_result.results {
                                sender
                                    .send_notice(
                                        &channel,
                                        format!("{author}: {} -- {service_name}", res.status),
                                    )
                                    .await
                                    .unwrap();
                            }
                        }
                        Err(err) => {
                            sender
                                .send_notice(&channel, format!("{author}: {err} -- {service_name}"))
                                .await
                                .unwrap();
                            error!("{} -- {}", err, service_name);
                        }
                    }
                }
            };
            return Ok(());
        }
        parsing::ParseResult::Unknown => return Ok(()),
    };
    let action_result = action_handle.await;
    if let Ok(action_result) = action_result {
        match action_result {
            Ok(action_result) => {
                if action_result.results.is_empty() {
                    sender
                        .send_notice(
                            &channel,
                            format!("{author}: Nothing to do -- {service_name}"),
                        )
                        .await
                        .unwrap();
                }

                for res in action_result.results {
                    sender
                        .send_notice(
                            &channel,
                            format!("{author}: {} -- {service_name}", res.status),
                        )
                        .await
                        .unwrap();
                }
            }
            Err(err) => {
                sender
                    .send_notice(&channel, format!("{author}: {err} -- {service_name}"))
                    .await
                    .unwrap();
                error!("{} -- {}", err, service_name);
            }
        }
    }
    Ok(())
}

type ActionResult = Result<
    tokio::task::JoinHandle<Result<icinga_api::api::ActionReply, icinga_api::icinga::error::Error>>,
    icinga_api::utils::Error,
>;

pub async fn highlight<HlFile, Author, Channel>(
    hl_file: HlFile,
    author: Author,
    channel: Channel,
    sender: &crate::sender::Sender,
) -> io::Result<()>
where
    HlFile: AsRef<Path>,
    Author: AsRef<str>,
    Channel: AsRef<str>,
{
    let hl = fs::read_to_string(&hl_file).unwrap_or_default();
    let author = author.as_ref();
    let channel = channel.as_ref();
    let remove_hl = hl.lines().filter(|line| *line != author);

    if remove_hl.clone().eq(hl.lines()) {
        if hl.is_empty() {
            fs::write(hl_file, author)?;
        } else {
            fs::write(hl_file, format!("{hl}\n{author}"))?;
        }
        sender
            .send_notice(&channel, format!("{author}: Highlight added for {author}"))
            .await
            .unwrap();
    } else {
        let new_hl = remove_hl.fold(String::new(), |s, line| {
            if s.is_empty() {
                String::from(line)
            } else {
                format!("{s}\n{line}")
            }
        });
        fs::write(hl_file, new_hl)?;
        sender
            .send_notice(
                &channel,
                format!("{author}: Highlight removed for {author}"),
            )
            .await
            .unwrap();
    }

    Ok(())
}

pub async fn help(author: &str, sender: &crate::sender::Sender) -> Result<(), irc::error::Error> {
    sender
        .send_notice(author, "Evolerto bot: bot to communicate with icinga.")
        .await?;
    sender.send_notice(author, "Commands:").await?;
    sender
        .send_notice(
            author,
            "`.ack <host>[!<service>[!<hoster>]] <comment>`: Acknowledge a problem",
        )
        .await?;
    sender.send_notice(author, "`.ackexp <host>[!<service>[!<hoster>]] <end_time> <comment>`: Temporarily acknowledge a problem").await?;
    sender.send_notice(author, "`.dt <host>[!<service>[!<hoster>]] <end_time> <comment>`: Schedule a downtime for an host or a service (start now)").await?;
    sender.send_notice(
        author,
        "`.dtat <host>[!<service>[!<hoster>]] <start_time> <end_time> <comment>`: Schedule a downtime for an host or a service at a future time (start_time in format <YYYY>-<mm>-<DD>T<HH>:<MM><+|-><HH>:<MM>)"
    ).await?;
    sender
        .send_notice(
            author,
            "`.com <host>[!<service>[!<hoster>]] <comment>`: Add a comment to an host or a service",
        )
        .await?;
    sender.send_notice(author, "`.rm <host>[!<service>[!<hoster>]]`: Remove all downtimes and ack on an host or a service").await?;
    sender
        .send_notice(
            author,
            "`.notif <true|false>`: Activate or deactivate notifications",
        )
        .await?;
    sender
        .send_notice(
            author,
            "`.hl`: Add yourself to those who should be notified of new alerts",
        )
        .await?;
    sender
        .send_notice(author, "`.help`: Show this help message")
        .await?;
    sender
        .send_notice(
            author,
            "Note: `<end_time>` is (in regex) '([1-9]+[wdhms])+' where w = weeks, d = days, h = hours, m = minutes and s = seconds",
        )
        .await
}

fn acknowledge<Connector>(
    p_res: i2api_utils::parsing::AckParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let action = api::AcknowledgeProblemAction::builder(filter, p_res.a_type)
        .filter_vars(p_res.filter_vars.into())
        .comment(p_res.comment)
        .author(author)
        .build();
    debug!("Ack: {:?}", action);
    Ok(tokio::spawn(async move {
        service.acknowledge_problem(&action).await
    }))
}

fn acknowledge_exp<Connector>(
    p_res: i2api_utils::parsing::AckExpParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let action = api::AcknowledgeProblemAction::builder(filter, p_res.a_type)
        .filter_vars(p_res.filter_vars.into())
        .comment(p_res.comment)
        .expiry(p_res.end_time)
        .author(author)
        .build();
    debug!("Ackexp: {:?}", action);
    Ok(tokio::spawn(async move {
        service.acknowledge_problem(&action).await
    }))
}

#[cfg(feature = "old_icinga")]
fn downtime<Connector>(
    p_res: i2api_utils::parsing::DtParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    use icinga_api::api::schedule_downtime_action::TargetType;

    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let comment = icinga_api::utils::NonEmptyStr::try_from(p_res.comment)?;
    let author = icinga_api::utils::NonEmptyStr::try_from(author).unwrap();
    let second_action = if matches!(p_res.a_type, TargetType::Host) {
        Some(
            api::ScheduleDowntimeActionBuilder::new(
                author.clone(),
                comment.clone(),
                p_res.end_time,
                filter.clone(),
                TargetType::Service,
            )
            .unwrap()
            .filter_vars(p_res.filter_vars.clone())
            .build(),
        )
    } else {
        None
    };
    let action = api::ScheduleDowntimeActionBuilder::new(
        author,
        comment,
        p_res.end_time,
        filter,
        p_res.a_type,
    )?
    .filter_vars(p_res.filter_vars)
    .build();
    debug!("Downtime: {:?}", action);
    Ok(tokio::spawn(async move {
        if let Some(second_action) = second_action {
            let _ = service.schedule_downtime(&second_action).await;
        }
        service.schedule_downtime(&action).await
    }))
}

#[cfg(not(feature = "old_icinga"))]
fn downtime<Connector>(
    p_res: i2api_utils::parsing::DtParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    use icinga_api::api::schedule_downtime_action::TargetType;

    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let comment = icinga_api::utils::NonEmptyStr::try_from(p_res.comment)?;
    let author = icinga_api::utils::NonEmptyStr::try_from(author).unwrap();
    let action = api::ScheduleDowntimeActionBuilder::new(
        author,
        comment,
        p_res.end_time,
        filter,
        p_res.a_type,
    )?
    .filter_vars(p_res.filter_vars.into())
    .all_services(matches!(p_res.a_type, TargetType::Host))
    .build();
    debug!("Downtime: {:?}", action);
    Ok(tokio::spawn(async move {
        service.schedule_downtime(&action).await
    }))
}

#[cfg(feature = "old_icinga")]
fn downtime_at<Connector>(
    p_res: i2api_utils::parsing::DtAtParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    use icinga_api::api::schedule_downtime_action::TargetType;

    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let comment = icinga_api::utils::NonEmptyStr::try_from(p_res.comment)?;
    let author = icinga_api::utils::NonEmptyStr::try_from(author).unwrap();
    let second_action = if matches!(p_res.a_type, TargetType::Host) {
        Some(
            api::ScheduleDowntimeActionBuilder::new(
                author.clone(),
                comment.clone(),
                p_res.end_time,
                filter.clone(),
                TargetType::Service,
            )?
            .filter_vars(p_res.filter_vars.clone())
            .start_time(p_res.start_time)?
            .build(),
        )
    } else {
        None
    };
    let action = api::ScheduleDowntimeActionBuilder::new(
        author,
        comment,
        p_res.end_time,
        filter,
        p_res.a_type,
    )?
    .filter_vars(p_res.filter_vars)
    .start_time(p_res.start_time)?
    .build();
    debug!("Downtime: {:?}", action);
    Ok(tokio::spawn(async move {
        if let Some(second_action) = second_action {
            let _ = service.schedule_downtime(&second_action).await;
        }
        service.schedule_downtime(&action).await
    }))
}

#[cfg(not(feature = "old_icinga"))]
fn downtime_at<Connector>(
    p_res: i2api_utils::parsing::DtAtParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    use icinga_api::api::schedule_downtime_action::TargetType;

    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let comment = icinga_api::utils::NonEmptyStr::try_from(p_res.comment)?;
    let author = icinga_api::utils::NonEmptyStr::try_from(author).unwrap();
    let action = api::ScheduleDowntimeActionBuilder::new(
        author,
        comment,
        p_res.end_time,
        filter,
        p_res.a_type,
    )?
    .filter_vars(p_res.filter_vars.into())
    .start_time(p_res.start_time)?
    .all_services(matches!(p_res.a_type, TargetType::Host))
    .build();
    debug!("Downtime: {:?}", action);
    Ok(tokio::spawn(async move {
        service.schedule_downtime(&action).await
    }))
}

fn comment<Connector>(
    p_res: i2api_utils::parsing::ComParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> ActionResult
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let action = api::AddCommentAction::builder(filter, p_res.a_type)
        .filter_vars(p_res.filter_vars.into())
        .comment(p_res.comment)
        .author(author)
        .build();
    debug!("Comment: {:?}", action);
    Ok(tokio::spawn(
        async move { service.add_comment(&action).await },
    ))
}

fn set_notification<Connector>(
    active: bool,
    service: Arc<icinga::Service<Connector>>,
) -> tokio::task::JoinHandle<Result<icinga_api::api::ActionReply, icinga_api::icinga::error::Error>>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    debug!("Notification: {:?}", active);
    tokio::spawn(async move { service.set_notification(active).await })
}

fn remove<Connector>(
    p_res: i2api_utils::parsing::RmParsed,
    service: Arc<icinga::Service<Connector>>,
    author: &str,
) -> Result<
    futures_util::future::TryJoinAll<
        tokio::task::JoinHandle<
            Result<icinga_api::api::ActionReply, icinga_api::icinga::error::Error>,
        >,
    >,
    icinga_api::utils::Error,
>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let filter = icinga_api::utils::NonEmptyStr::try_from(p_res.filter)?;
    let filter_vars = HashMap::from(p_res.filter_vars);
    let dt_type = match p_res.a_type {
        ObjectType::Service => p_res.a_type,
        ObjectType::Host => ObjectType::Downtime,
        _ => unreachable!(),
    };
    let dt_action = api::RemoveDowntimeActionBuilder::new(filter.clone(), dt_type.try_into()?)
        .filter_vars(filter_vars.clone())
        .author(author)
        .build();
    let dt_handle = {
        let service = service.clone();
        tokio::spawn(async move { service.remove_downtime(&dt_action).await })
    };
    let ack_action1 =
        api::RemoveAcknowledgementActionBuilder::new(filter.clone(), p_res.a_type.try_into()?)
            .filter_vars(filter_vars.clone())
            .author(author)
            .build();
    let ack_handle1 = {
        let service = service.clone();
        tokio::spawn(async move { service.remove_acknowledgement(&ack_action1).await })
    };
    let join_handle = match p_res.a_type {
        ObjectType::Service => try_join_all([dt_handle, ack_handle1]),
        ObjectType::Host => {
            let ack_action2 = api::RemoveAcknowledgementActionBuilder::new(
                filter,
                icinga_api::api::remove_acknowledgement_action::TargetType::Service,
            )
            .filter_vars(filter_vars)
            .author(author)
            .build();
            let ack_handle2 =
                tokio::spawn(async move { service.remove_acknowledgement(&ack_action2).await });
            try_join_all([dt_handle, ack_handle1, ack_handle2])
        }
        _ => unreachable!(),
    };
    Ok(join_handle)
}
