use alloc::format;

use super::parsing::ParseActionError;
use super::sender;

pub(super) async fn handle_api_errors(
    error: icinga_api::utils::Error,
    original_command: &str,
    sender: &sender::Sender,
    channel: &str,
) -> Result<(), irc::error::Error> {
    use icinga_api::utils::{Error, RequestBuilderError};

    tracing::warn!(command = original_command, "An error happened: {error:?}");
    match error {
        Error::EmptyStrError => Err(error).expect("There shouldn't be any empty string at this point."),
        Error::InvalidActionTypeError(error) => Err(error).expect("Action types should already be limited to the correct ones"),
        Error::RequestBuilderError(error) => {
            match error {
                RequestBuilderError::InvalidActionType(error) => Err(error).expect("Action types should already be limited to the correct ones"),
                RequestBuilderError::StartAfterEnd { end, start } => sender
                    .send_notice(&channel, format!("The provided start time {start} happens AFTER the end time {end} which really shouldn't be possible. Please report how you did this."))
                    .await,
                RequestBuilderError::EndBeforeStart { end, start } => sender
                    .send_notice(&channel, format!("The provided end time {end} happens BEFORE the start time {start} which really shouldn't be possible. Please report how you did this."))
                    .await,
            }
        },
    }
}    

pub(super) async fn handle_parsing_errors(error: ParseActionError, original_command: &str, channel: &str, sender: &sender::Sender) {
    match error {
        ParseActionError::ParseEndTime(_) => {
            tracing::info!(
                command = original_command,
                "Received a command with an invalid end time."
            );
            sender
                .send_notice(
                    channel,
                    "Invalid end time: Specify end time with this format X[m|h|d|w]",
                )
                .await
                .unwrap();
        }
        ParseActionError::IncorrectStartTime(_) => {
            tracing::info!(
                command = original_command,
                "Received a command with an invalid start time."
            );
            sender
                .send_notice(channel, "Invalid start time: Start time should be in Iso8601 format with minute precision and timezone.")
                .await
                .unwrap();
        }
        ParseActionError::TargetParse(err) => {
            tracing::info!(
                command = original_command,
                "Received a command with an unparseable target (error: {err})"
            );
            sender
                .send_notice(
                    channel,
                    "Invalid target: This target couldn't be parsed, if it was a valid target, please report as a bug."
                )
                .await
                .unwrap();
        }
        ParseActionError::MissingTarget => {
            tracing::info!(
                command = original_command,
                "Received a command needing a target but missing one."
            );
            sender
                .send_notice(
                    channel,
                    "Missing target: This command needs a target to be valid."
                )
                .await
                .unwrap();
        }
        ParseActionError::GenericParse => {
            tracing::warn!(
                command = original_command,
                "Received invalid command : {original_command}"
            );
            sender
                .send_notice(
                    channel,
                    format!("Your command: '.{original_command}' wasn't understood."),
                )
                .await
                .unwrap();
        }
    }
}
