use alloc::format;
use alloc::string::String;
use alloc::sync::Arc;
use core::fmt::Debug;

use futures_util::future::join_all;
use hyper::client::connect::Connect;
use i2api_utils::parsing;
use icinga_api::icinga;
use time_tz::OffsetDateTimeExt;
use tracing::{debug, debug_span, instrument};
use tracing_futures::Instrument;

use crate::{action, sender, HL_FILE};

mod error_handling;

#[instrument]
async fn send_end_time(
    parse_result: &parsing::ParseResult,
    target: &str,
    author: &str,
    sender: &sender::Sender,
) -> Result<(), irc::error::Error> {
    let system_tz = time_tz::system::get_timezone().unwrap_or(time_tz::timezones::db::UTC);

    let end_time = match parse_result {
        parsing::ParseResult::AckExp(parse_result) => parse_result.end_time.to_timezone(system_tz),
        parsing::ParseResult::Dt(parse_result) => parse_result.end_time.to_timezone(system_tz),
        parsing::ParseResult::DtAt(parse_result) => parse_result.end_time.to_timezone(system_tz),
        _ => return Ok(()),
    };

    let year = end_time.year();
    let month = end_time.month() as u8;
    let day = end_time.day();
    let hour = end_time.hour();
    let minute = end_time.minute();
    let second = end_time.second();
    let offset = end_time.offset();

    let message = match parse_result {
        parsing::ParseResult::AckExp(_) => format!("{author}: This acknowledgement will last until {year:04}-{month:02}-{day:02} {hour:02}:{minute:02}:{second:02}{offset}"),
        parsing::ParseResult::Dt(_) |
        parsing::ParseResult::DtAt(_) => format!("{author}: This downtime will last until {year:04}-{month:02}-{day:02} {hour:02}:{minute:02}:{second:02}{offset}"),
        _ => return Ok(()),
    };

    sender.send_notice(target, message).await
}

#[instrument]
pub(crate) fn read_command<Connector>(
    services: &[Arc<icinga::Service<Connector>>],
    mut input_stream: irc::client::ClientStream,
    sender: &Arc<sender::Sender>,
    rt: &tokio::runtime::Runtime,
) where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    use futures_util::stream::StreamExt;
    rt.block_on(async {
        while let Some(message) = input_stream.next().await.transpose().unwrap() {
            let author = Arc::new(String::from(message.source_nickname().unwrap_or_default()));
            let response_target = message.response_target().map(String::from);
            if let irc::client::prelude::Command::PRIVMSG(channel, message_content) =
                message.command
            {
                match message_content.strip_prefix('.') {
                    None => continue,
                    Some(buf) => {
                        debug!(command = buf, "Received command : {}", buf);
                        let parse_result = {
                            let mut command = buf.split_whitespace();
                            parsing::parse_command(command.next().unwrap_or_default())
                                .parse(command)
                        };
                        let channel = Arc::new(response_target.unwrap_or(channel));
                        act_on_command(parse_result, buf, author, channel, services, sender).await;
                    }
                }
            }
        }
    });
}

async fn act_on_command<Connector>(
    parsed_action: Result<parsing::ParseResult, parsing::ParseActionError>,
    original_command: &str,
    author: Arc<String>,
    channel: Arc<String>,
    services: &[Arc<icinga::Service<Connector>>],
    sender: &Arc<sender::Sender>,
) where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    match parsed_action {
        Ok(parsing::ParseResult::Highlight) => {
            action::highlight(
                HL_FILE.get().unwrap(),
                author.as_ref(),
                channel.as_ref(),
                sender,
            )
            .await
            .unwrap();
        }
        Ok(parsing::ParseResult::Help) => {
            action::help(&author, sender).await.unwrap();
        }
        Ok(parsed_action) => {
            let ret = join_all(services.iter().cloned().map(|service| {
                let parse_result = parsed_action.clone();
                let author = Arc::clone(&author);
                let channel = Arc::clone(&channel);
                let sender = Arc::clone(sender);
                tokio::spawn(async move {
                    action::do_actions(&service, parse_result, &author, &channel, &sender)
                        .instrument(debug_span!("action"))
                        .await
                })
            }))
            .await;

            if let Some(err) = ret.into_iter().flatten().find_map(Result::err) {
                error_handling::handle_api_errors(err, original_command, sender, &channel)
                    .await
                    .unwrap();
            }
            send_end_time(&parsed_action, &channel, &author, sender)
                .await
                .unwrap();
        }
        Err(error) => error_handling::handle_parsing_errors(error, original_command, &channel, sender).await,
    }
}
