use alloc::boxed::Box;
use alloc::format;
use core::fmt::Debug;
use core::sync::atomic::{AtomicBool, Ordering};

use anyhow::Context;
use futures_util::future::join;
use hyper::client::connect::Connect;
use icinga_api::{api, icinga};
use once_cell::sync::Lazy;
use tracing::{debug, instrument};

use crate::{fs, sender, thread, HashSet, Never, HL_FILE, PRIMARY_CHANNEL};

const SERVICE_FILTER: &str = "service.state>=2 && service.downtime_depth == 0 && service.acknowledgement ==0 && service.enable_notifications == 1 && host.state == 0 && service.state_type == 1";
const HOST_FILTER: &str = "(host.state == 1 || host.state == 2) && host.downtime_depth == 0 && host.acknowledgement == 0 && host.enable_notifications == 1 && host.state_type == 1";
const _TIME_AFTER_RESTART: u32 = 4_u32;
const SLEEP_BETWEEN_LOOPS: core::time::Duration = core::time::Duration::from_secs(3);

static SERVICE_QUERY: Lazy<api::QueryRequest> = Lazy::new(build_service_query);
static HOST_QUERY: Lazy<api::QueryRequest> = Lazy::new(build_host_query);

fn build_service_query() -> api::QueryRequest {
    api::QueryRequest::builder()
        .attrs(["display_name"])
        .filter(SERVICE_FILTER)
        .build()
}

fn build_host_query() -> api::QueryRequest {
    api::QueryRequest::builder()
        .attrs(["display_name"])
        .joins(["host.name"])
        .filter(HOST_FILTER)
        .build()
}

#[instrument(skip(service))]
async fn list_services_hs<Connector>(
    service: &icinga::Service<Connector>,
) -> Result<impl Iterator<Item = Box<str>>, icinga_api::icinga::error::Error>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let query = Lazy::force(&SERVICE_QUERY);
    debug!("Query service: {query:?}");

    let ret = service
        .query_service(query)
        .await?
        .results
        .into_iter()
        .map(|res| res.name);
    Ok(ret)
}

#[instrument(skip(service))]
async fn list_hosts_hs<Connector>(
    service: &icinga::Service<Connector>,
) -> Result<impl Iterator<Item = Box<str>>, icinga_api::icinga::error::Error>
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let query = Lazy::force(&HOST_QUERY);

    debug!("Query host: {query:?}");
    let ret = service
        .query_hosts(query)
        .await?
        .results
        .into_iter()
        .map(|res| res.name);
    Ok(ret)
}

async fn send_to_hl_and_primary<S: AsRef<str>>(message: S, irc_sender: &sender::Sender) {
    async fn internal_send_to_hl_and_primary(message: &str, irc_sender: &sender::Sender) {
        let hl = fs::read_to_string(HL_FILE.get().unwrap()).unwrap_or_default();
        for nick in hl.lines() {
            if let Err(error) = irc_sender.send_notice(nick, message).await {
                tracing::warn!("{error}");
            }
        }
        if let Err(error) = irc_sender
            .send_notice(PRIMARY_CHANNEL.get().unwrap(), message)
            .await
        {
            tracing::warn!("{error}");
            irc_sender
                .send_join(PRIMARY_CHANNEL.get().unwrap())
                .await
                .unwrap();
        }
    }
    internal_send_to_hl_and_primary(message.as_ref(), irc_sender).await;
}

async fn _check_for_recent_restart<Connector>(service: &icinga::Service<Connector>) -> bool
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let status = match service.get_application_status().await {
        Ok(status) => status,
        Err(err) => {
            tracing::warn!("Error retrieving status : {err}");
            return true;
        }
    };
    let Some(status) = status.results.first() else { return true; };

    let service_start = status.status.icingaapplication.app.program_start();
    let current = time::Duration::seconds(time::OffsetDateTime::now_utc().unix_timestamp());

    let calculated = current - time::Duration::seconds_f64(service_start);
    tracing::debug!(
        icinga_start = service_start,
        calculated_time = calculated.as_seconds_f64(),
        "Calculating time since icinga start"
    );
    calculated < time::Duration::minutes(_TIME_AFTER_RESTART.into())
}

fn treat_returned_alerts<Services, Hosts, F>(
    services: Services,
    hosts: Hosts,
    current_alertes: &HashSet<Box<str>>,
    last_empty: bool,
    handler: F,
) -> (Option<HashSet<Box<str>>>, bool)
where
    Services: IntoIterator<Item = Box<str>>,
    Hosts: IntoIterator<Item = Box<str>>,
    F: FnMut(&str),
{
    #[cold]
    const fn no_alerts() -> bool {
        true
    }

    fn treatment<F>(
        new_alertes: HashSet<Box<str>>,
        current_alertes: &HashSet<Box<str>>,
        last_empty: bool,
        mut handler: F,
    ) -> (Option<HashSet<Box<str>>>, bool)
    where
        F: FnMut(&str),
    {
        // Double check that there are no alertes.
        if !last_empty && new_alertes.is_empty() {
            return (None, no_alerts());
        }
        for alerte in new_alertes.difference(current_alertes) {
            handler(alerte);
        }
        (Some(new_alertes), false)
    }
    let new_alertes: HashSet<Box<str>> = services.into_iter().chain(hosts).collect();
    treatment(new_alertes, current_alertes, last_empty, handler)
}

#[cold]
async fn _treat_recent_restart(sender: &sender::Sender, minutes: u32) {
    tracing::info!(
        "Icinga seems to have recently restarted, sleeping for a bit to avoid useless noise."
    );
    send_to_hl_and_primary(
        "Icinga has restarted, no alerte will be sent for some time (so as not to spam you)",
        sender,
    )
    .await;
    // Icinga has some time between API availability and ack and dt being reapplied, wait a bit before continuing if it has restarted recently.
    tokio::time::sleep(core::time::Duration::from_secs(60_u64).saturating_mul(minutes)).await;
}

#[instrument(level = "trace")]
pub(crate) fn read_icinga<Connector>(
    service: &icinga::Service<Connector>,
    irc_sender: &sender::Sender,
    rt: &tokio::runtime::Runtime,
) -> Never
where
    Connector: Connect + Clone + Send + Sync + Debug + 'static,
{
    let mut current_alertes: HashSet<Box<str>> = HashSet::default();
    let last_empty = AtomicBool::new(false);
    loop {
        let handle_services = async {
            list_services_hs(service).await.with_context(|| {
                format!(
                    "Failed to recover the list of critical services from {}",
                    service.url()
                )
            })
        };
        let handle_hosts = async {
            list_hosts_hs(service).await.with_context(|| {
                format!(
                    "Failed to recover the list of critical hosts from {}",
                    service.url()
                )
            })
        };
        let (tmp_services, tmp_hosts) = rt.block_on(join(handle_services, handle_hosts));
        // if rt.block_on(_check_for_recent_restart(service)) {
        //     rt.block_on(_treat_recent_restart(irc_sender, _TIME_AFTER_RESTART));
        //     last_empty.store(false, Ordering::Release);
        //     continue;
        // }
        match (tmp_hosts, tmp_services) {
            (Ok(tmp_hosts), Ok(tmp_services)) => {
                let (new_alertes, tmp_empty) = treat_returned_alerts(
                    tmp_services,
                    tmp_hosts,
                    &current_alertes,
                    last_empty.load(Ordering::Acquire),
                    |alerte| rt.block_on(send_to_hl_and_primary(alerte, irc_sender)),
                );
                last_empty.store(tmp_empty, Ordering::Release);
                if let Some(new_alertes) = new_alertes {
                    current_alertes = new_alertes;
                }
            }
            (Err(error), Ok(_)) | (Ok(_), Err(error)) => {
                tracing::error!("{error}");
            }
            (Err(error1), Err(error2)) => {
                tracing::error!("{error1}");
                tracing::error!("{error2}");
            }
        }
        thread::sleep(SLEEP_BETWEEN_LOOPS);
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn has_alerts() {
        const LAST_EMPTY: bool = false;

        let alertes_s1 = ["tmp", "tmp2", "tmp3"].map(Box::from);
        let alertes_s2 = ["tmp4"].map(Box::from);

        let expected_alertes = {
            let mut tmp = HashSet::default();
            tmp.extend(alertes_s1.clone());
            tmp.extend(alertes_s2.clone());
            tmp
        };

        let (returned_alertes, current_empty, ran_handler) = {
            let mut ran_handler = false;
            let current_alertes = HashSet::default();

            let tmp =
                treat_returned_alerts(alertes_s1, alertes_s2, &current_alertes, LAST_EMPTY, |_| {
                    ran_handler = true;
                });
            (tmp.0, tmp.1, ran_handler)
        };

        assert!(!current_empty);
        assert!(ran_handler);
        assert_eq!(returned_alertes, Some(expected_alertes));
    }

    #[test]
    fn no_alerts() {
        const ALERTES: [Box<str>; 0] = [];
        const LAST_EMPTY: bool = false;

        let current_alertes = HashSet::default();
        let mut ran_handler = false;

        let (returned_alertes, current_empty) =
            treat_returned_alerts(ALERTES, ALERTES, &current_alertes, LAST_EMPTY, |_| {
                ran_handler = true;
            });
        let ran_handler = ran_handler;

        assert!(current_empty);
        assert!(!ran_handler);
        assert_eq!(returned_alertes, None);
    }
}
