#![cfg_attr(feature = "use_nightly", feature(never_type))]
#![forbid(unsafe_code)]
#![warn(clippy::pedantic)]
#![no_std]
extern crate alloc;
extern crate std;

use alloc::format;
use alloc::string::String;
use alloc::vec::Vec;
use std::ffi::OsString;
use std::io;
use std::path::PathBuf;
use std::sync::Arc;
pub(crate) use std::{fs, thread};

use anyhow::Result;
use clap::{Arg, Command};
use commands::read_command;
use hyper::client::HttpConnector;
use hyper_rustls::HttpsConnectorBuilder;
use hyper_trust_dns::TrustDnsResolver;
use i2api_utils::conf;
use icinga_api::icinga;
use list_alerts::read_icinga;
use once_cell::sync::OnceCell;
pub(crate) use rustc_hash::FxHashSet as HashSet;
use rustls::ClientConfig;

mod action;
mod commands;
mod list_alerts;
mod sender;
mod tls;

#[cfg(feature = "use_nightly")]
pub(crate) type Never = !;
#[cfg(not(feature = "use_nightly"))]
pub(crate) enum Never {}

static PRIMARY_CHANNEL: OnceCell<String> = OnceCell::new();
static HL_FILE: OnceCell<OsString> = OnceCell::new();

fn setup_logger() {
    use tracing_subscriber::prelude::*;

    let journald_logger = tracing_journald::layer().ok();
    let fmt_logger = if journald_logger.is_some() {
        None
    } else {
        Some(tracing_subscriber::fmt::layer().with_writer(io::stderr))
    };
    tracing_subscriber::registry()
        .with(tracing_subscriber::EnvFilter::from_default_env())
        .with(journald_logger)
        .with(fmt_logger)
        .init();
}

fn create_cli() -> clap::Command {
    Command::new("Icinga bot")
        .version(clap::crate_version!())
        .author(clap::crate_authors!("\n"))
        .about(clap::crate_description!())
        .arg(
            Arg::new("config")
                .short('c')
                .long("config")
                .value_name("CONF_FILE")
                .help("Sets a custom config file")
                .num_args(1)
                .default_value("conf.toml")
                .value_parser(clap::builder::ValueParser::path_buf()),
        )
        .arg(
            Arg::new("irc_conf")
                .long("irc-config")
                .value_name("IRC_CONF")
                .help("Config file for IRC")
                .num_args(1)
                .default_value("irc.toml")
                .value_parser(clap::builder::ValueParser::path_buf()),
        )
        .arg(
            Arg::new("master")
                .short('m')
                .long("master")
                .value_name("SERVICE")
                .help("Set the icinga node that will be queried for informations")
                .num_args(1)
                .value_parser(clap::builder::ValueParser::string()),
        )
        .arg(
            Arg::new("hl_file")
                .long("hl-file")
                .value_name("HL_FILE")
                .help("Sets the file in which to save current highlight")
                .num_args(1)
                .default_value("hl.tmp")
                .value_parser(clap::builder::ValueParser::os_string()),
        )
}

fn password_from_config(conf: &conf::CService) -> String {
    conf.password_file
        .as_ref()
        .and_then(|password_file| {
            let password_file = PathBuf::from(password_file);
            std::fs::read_to_string(password_file).ok()
        })
        .or_else(|| conf.password.clone())
        .expect("A password has to be provided, either through `password` or a file (`password_file`)")
}

fn icinga_from_conf<Connector>(client: hyper::Client<Connector>, name: String, conf: conf::CService) -> Option<icinga::Service<Connector>>
where Connector: hyper::client::connect::Connect + Clone + Send + Sync + core::fmt::Debug + 'static,
{
    let password = password_from_config(&conf);

    icinga::ServiceBuilder::new(client)
        .url(conf.url)
        .auth(conf.username, password)
        .name(name)
        .build()
}

#[allow(clippy::too_many_lines)]
fn main() -> Result<()> {
    setup_logger();

    let args = create_cli().get_matches();

    let conf_file = fs::read_to_string(args.get_one::<PathBuf>("config").unwrap())?;
    let irc_conf_file = fs::read_to_string(args.get_one::<PathBuf>("irc_conf").unwrap())?;
    let config: conf::Config = toml::from_str(&conf_file)?;
    PRIMARY_CHANNEL
        .set(config.primary_channel.unwrap_or_default())
        .unwrap();
    HL_FILE
        .set(args.get_one::<OsString>("hl_file").unwrap().clone())
        .unwrap();
    let throttling = config.throttling.unwrap();
    let master_name = args
        .get_one::<String>("master")
        .cloned()
        .or(config.master)
        .expect("Master must be defined");
    let master = config.hosts.get(&master_name).cloned().unwrap();
    let client = {
        let resolver = TrustDnsResolver::from_system_conf();
        let client_config = ClientConfig::builder()
            .with_safe_defaults()
            .with_custom_certificate_verifier(Arc::new(tls::NoVerifier))
            .with_no_client_auth();
        let internal_connector = {
            let mut tmp = HttpConnector::new_with_resolver(resolver);
            tmp.enforce_http(false);
            tmp.set_connect_timeout(config.http_connect_timeout);
            tmp
        };
        let connector = HttpsConnectorBuilder::new()
            .with_tls_config(client_config)
            .https_or_http()
            .enable_http1()
            .wrap_connector(internal_connector);
        hyper::Client::builder()
            .retry_canceled_requests(false)
            .build(connector)
    };

    let icingas = {
        let client = client.clone();
        config
            .hosts
            .into_iter()
            .map(|(srv, conf)| {
                icinga_from_conf(client.clone(), srv, conf)
                    .unwrap()
            })
            .map(Arc::new)
            .collect()
    };
    let icinga_master = {
        icinga_from_conf(client.clone(), master_name, master)
            .unwrap()
    };

    let rt = Arc::new(
        tokio::runtime::Builder::new_multi_thread()
            .enable_all()
            .build()?,
    );

    let mut irc = rt
        .block_on(irc::client::Client::from_config(toml::from_str(
            &irc_conf_file,
        )?))
        .unwrap();
    irc.identify().unwrap();
    let sender = Arc::new(sender::Sender::new(
        irc.sender(),
        throttling.max_messages,
        core::time::Duration::from_secs(throttling.window.get() as u64),
    ));

    thread::scope(|scope| {
        let _read_handle = {
            let icinga = icinga_master;
            let sender = sender.clone();
            let rt = rt.clone();
            thread::Builder::new()
                .name(String::from("read_icinga"))
                .spawn_scoped(scope, move || read_icinga(&icinga, &sender, &rt))
                .unwrap()
        };
        let outgoing = irc.outgoing().unwrap();
        let _command_handle = {
            let icingas: Vec<_> = icingas;
            let rt = rt.clone();
            let irc_stream = irc.stream().unwrap();
            thread::Builder::new()
                .name(String::from("command"))
                .spawn_scoped(scope, move || {
                    read_command(&icingas, irc_stream, &sender, &rt);
                })
                .unwrap()
        };
        core::mem::drop(rt.block_on(outgoing));
    });
    Ok(())
}
