use alloc::string::ToString;
use core::fmt::Display;
use core::ops::Div;

use irc::client::prelude::*;
use irc::proto::mode::ModeType;

use crate::String;

pub struct Sender {
    sender: irc::client::Sender,
    rate_limiter: governor::RateLimiter<
        String,
        governor::state::keyed::DefaultKeyedStateStore<String>,
        governor::clock::DefaultClock,
    >,
    max_jitter: governor::Jitter,
}

impl core::fmt::Debug for Sender {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.debug_struct("Sender")
            .field("sender", &self.sender)
            .finish()
    }
}

impl Sender {
    pub fn new(
        sender: irc::client::Sender,
        max_messages: core::num::NonZeroU32,
        interval: core::time::Duration,
    ) -> Self {
        let quota = governor::Quota::with_period(interval)
            .expect("Interval should be > 0")
            .allow_burst(max_messages);
        Self {
            sender,
            rate_limiter: governor::RateLimiter::keyed(quota),
            max_jitter: governor::Jitter::up_to(interval.div(3)),
        }
    }

    #[allow(dead_code)]
    pub async fn send<M: Into<irc::client::prelude::Message>>(
        &self,
        msg: M,
    ) -> irc::error::Result<()> {
        self.rate_limiter
            .until_key_ready_with_jitter(&String::default(), self.max_jitter)
            .await;
        self.sender.send(msg)
    }

    #[allow(dead_code)]
    pub async fn send_mode<S, T>(&self, target: S, modes: &[Mode<T>]) -> irc::error::Result<()>
    where
        S: Display,
        T: ModeType,
    {
        self.rate_limiter
            .until_key_ready_with_jitter(&target.to_string(), self.max_jitter)
            .await;
        self.sender.send_mode(target, modes)
    }

    pub async fn send_join(&self, chanlist: impl Display) -> irc::error::Result<()> {
        self.rate_limiter
            .until_key_ready_with_jitter(&chanlist.to_string(), self.max_jitter)
            .await;
        self.sender.send_join(chanlist)
    }

    pub async fn send_notice(
        &self,
        target: impl Display,
        message: impl Display,
    ) -> irc::error::Result<()> {
        self.rate_limiter
            .until_key_ready_with_jitter(&target.to_string(), self.max_jitter)
            .await;
        self.sender.send_notice(target, message)
    }
}
